# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The module contains version information about the tosfs package."""

#
# Generated by scripts/generate_version.sh
#
class Version:
    """The version information about the tosfs package."""
    version = "2024.11.4"
    revision = "732d935e9462ea97b3b92723005a6a16bd59ba87"
    user = "bytedance"
    date = "Tue Nov 26 19:45:14 CST 2024"
    repo_url = "https://github.com/fsspec/tosfs.git"
